# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# chATLAS_Embed is free software; you can redistribute it and/or modify
# it under the terms of the Apache 2.0 license; see LICENSE file for more details.
# `chATLAS_Embed/VectorStoreCreators.py`
"""
A list of scientific stop words to use and exclude for a physics HEP Area

Also contains some gathered english stop words from various other areas as redundency to pg english dict

NOTE DO NOT INCLUDE WORDS WITH ' in - this breaks the SQL (it is not well made) and is unecessary as we split on
    punctuation anyway (so just include stem).
"""


def get_physics_stopwords():
    """
    Returns a list of common stop words for HEP physics documents.
    Categorized for easy modification.
    """
    # Units and their variations
    units = [
        "mev",
        "gev",
        "tev",
        "pev",
        "ev",  # Energy units
        "mm",
        "cm",
        "m",
        "km",  # Length units
        "ms",
        "ns",
        "ps",
        "fs",  # Time units
        "pb",
        "fb",
        "nb",  # Cross section units
        "rad",
        "mrad",  # Angular units
    ]

    # Common prefixes that appear frequently but add little search value
    prefixes = [
        "pre",
        "post",
        "sub",
        "super",
        "anti",
        "non",
        "semi",
        "ultra",
    ]

    # Common technical terms that are too generic to be useful search terms
    generic_technical = [
        "data",
        "run",
        "runs",
        "event",
        "events",
        "value",
        "values",
        "error",
        "errors",
        "measurement",
        "measurements",
        "analysis",
        "analyses",
        "cut",
        "cuts",
        "plot",
        "plots",
        "fig",
        "figure",
        "figures",
        "ref",
        "refs",
        "reference",
        "references",
        "sys",
        "syst",
        "systematic",
        "systematics",
        "stat",
        "statistical",
        "eq",
        "eqn",
        "equation",
        "equations",
    ]

    # Common ATLAS-specific terms that might be too frequent to be useful
    atlas_generic = [
        "atlas",
        "cern",
        "lhc",
        "trigger",
        "daq",
        "detector",
        "beam",
        "beams",
        "collision",
        "collisions",
        "luminosity",
        "period",
        "calibration",
        "reconstruction",
    ]

    # Common mathematical/statistical terms that are too generic
    math_stats = [
        "mean",
        "std",
        "dev",
        "min",
        "max",
        "avg",
        "average",
        "sum",
        "total",
        "calc",
        "calculation",
        "function",
        "parameter",
        "parameters",
        "variable",
        "variables",
        "correlation",
        "fit",
        "fits",
        "fitted",
    ]

    # Common document structure words
    document_structure = [
        "section",
        "subsection",
        "chapter",
        "appendix",
        "table",
        "caption",
    ]

    releases = [
        "rel12",
        "rel13",
        "rel14",
        "rel15",
        "rel16",
        "rel17",
        "rel18",
        "r12",
        "r13",
        "r14",
        "r15",
        "r16",
        "r17",
        "r18",
        "release15",
        "release16",
        "mc08",
        "mc09",
        "mc10",
        "mc11",
        "mc12",
        "mc14",
        "mc15",
        "dc14",
    ]

    years = [
        "run1",
        "ls1",
        "2009",
        "2010",
        "2011",
        "2012",
        "2013",
        "2014",
        "7tev",
        "8tev",
    ]

    things = [
        "dpd",
        "d2pd",
        "d3pd",
        "minimumbias13",
        "hcw09",
        "hcw13",
        "hcw14",
        "hadroniccalibration1",
        "hadroniccalibrationworkshop09",
        "pasttaucpworkshops",
        "muonboydocumentation",
    ]

    equation_leftovers = ["-", "---", "+", "{", "}", ""]

    version_patterns = [
        "v1",
        "v2",
        "v3",
        "v1.0",
        "v2.0",
        "v3.0",
        "version",
    ]

    # Computing and data processing terms
    computing_terms = [
        "cpu",
        "gpu",
        "ram",
        "memory",
        "disk",
        "storage",
        "buffer",
        "cache",
        "compression",
        "compressed",
        "uncompress",
        "uncompressed",
        "bandwidth",
        "rate",
        "throughput",
        "input",
        "output",
        "i/o",
        "db",
        "database",
        "dataflow",
        "workflow",
    ]

    # Size and speed units
    size_speed_units = [
        "kb",
        "mb",
        "gb",
        "tb",
        "pb",
        "hz",
        "khz",
        "mhz",
        "ghz",
        "mb/s",
        "gb/s",
    ]

    # Common infrastructure terms
    infrastructure_terms = [
        "tier",
        "tier0",
        "tier1",
        "tier2",
        "t0",
        "t1",
        "t2",
        "sfo",
        "daq",
        "hlt",
        "grid",
        "cloud",
        "cluster",
        "node",
        "server",
        "service",
        "api",
        "interface",
    ]

    # Document status and metadata
    status_terms = [
        "draft",
        "final",
        "review",
        "approved",
        "pending",
        "todo",
        "done",
        "status",
        "state",
        "updated",
        "modified",
        "created",
        "attached",
        "latest",
        "current",
        "previous",
        "next",
    ]

    # Common technical actions
    action_terms = [
        "process",
        "processed",
        "processing",
        "execute",
        "executed",
        "executing",
        "run",
        "running",
        "start",
        "stop",
        "restart",
        "initialize",
        "configure",
        "setup",
        "install",
        "update",
        "upgrade",
        "downgrade",
    ]

    # Common file operations
    file_terms = [
        "file",
        "files",
        "directory",
        "folder",
        "path",
        "location",
        "store",
        "stored",
        "save",
        "saved",
        "load",
        "loaded",
        "backup",
        "restore",
        "archive",
        "compress",
        "decompress",
        "zip",
        "unzip",
    ]

    # Time-related terms
    time_terms = [
        "time",
        "timestamp",
        "date",
        "period",
        "duration",
        "interval",
        "frequency",
        "rate",
        "schedule",
        "timing",
    ]

    # Additional mathematical/programming terms
    additional_technical = [
        "flag",
        "flags",
        "option",
        "options",
        "config",
        "configuration",
        "setting",
        "settings",
        "property",
        "properties",
        "attribute",
        "attributes",
        "method",
        "methods",
        "function",
        "functions",
        "routine",
        "routines",
        "algorithm",
        "algorithms",
    ]

    pure_numbers = [f"{i}" for i in range(-1000, 1000)]

    build_terms = [
        "cmake",
        "make",
        "build",
        "compile",
        "configure",
        "configuration",
        "layout",
        "package",
        "dependency",
        "dependencies",
        "install",
        "setup",
        "ide",
    ]

    # ATLAS detector specific terms
    detector_terms = [
        "indet",
        "lar",
        "tile",
        "mdt",
        "trt",
        "sct",
        "rpvll",
        "desdm",
        "desd",
        "cesium",
        "laser",
    ]

    # Common URL and documentation patterns
    doc_patterns = [
        "twiki",
        "cern.ch",
        "viewauth",
        "atlascomputing",
        "homepage",
        "doxygen",
        "document",
        "documentation",
    ]

    # Time and date related
    time_date = [
        "january",
        "february",
        "march",
        "april",
        "may",
        "june",
        "july",
        "august",
        "september",
        "october",
        "november",
        "december",
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
        "sunday",
    ]

    # Common numerical patterns (to be added to equation_leftovers) - Should now also be caught by REGEX
    numerical_patterns = [
        "-0",
        "-1",
        "-2",
        "-3",
        "-4",
        "-01",
        "-02",
        "-03",
        "-04",
        "2e",
        "1e",
        "3e",
        "4e",
    ]

    # Common file paths and patterns
    file_paths = [
        "ofl",
        "bin",
        "lib",
        "include",
        "src",
        "test",
        "doc",
        "docs",
        "indet",
        "lar",
        "tile",
        "trt",
        "badchannels",
        "channeloffset",
    ]

    # Common status and quality terms
    status_terms = [
        "valid",
        "invalid",
        "complete",
        "incomplete",
        "ready",
        "pending",
        "done",
        "todo",
        "available",
        "unavailable",
        "usual",
        "unusual",
        "normal",
        "abnormal",
        "lock",
        "locked",
        "unlock",
        "unlocked",
    ]

    file_types = [
        "png",
        "csv",
        "png",
        "csv",
        "jpg",
        "jpeg",
        "gif",
        "bmp",
        "txt",
        "doc",
        "docx",
        "pdf",
        "xls",
        "xlsx",
    ]

    other_english_stop_words = [
        "box",
        "phys",
        # Common verbs
        "are",
        "be",
        "was",
        "were",
        "being",
        "could",
        "would",
        "will",
        "want",
        "use",
        "used",
        "using",
        "done",
        "going",
        "get",
        "got",
        "need",
        "needed",
        "help",
        "helped",
        # Common prepositions/articles
        "due",
        "for",
        "from",
        "to",
        "with",
        "around",
        "into",
        "of",
        "by",
        # Time-related words
        "today",
        "tomorrow",
        "next",
        "last",
        "day",
        "week",
        "april",
        # Common adjectives
        "ready",
        "possible",
        "final",
        "still",
        "well",
        "far",
        # Other common words
        "also",
        "otherwise",
        "follow",
        "left",
        "right",
        "top",
        "end",
        "part",
        "side",
        "note",
        "inform",
        "keep",
        "put",
    ]

    github_english_stop_words = [
        "10",
        "39",
        "a",
        "able",
        "ableabout",
        "about",
        "above",
        "abroad",
        "abst",
        "accordance",
        "according",
        "accordingly",
        "across",
        "act",
        "actually",
        "ad",
        "added",
        "adj",
        "adopted",
        "ae",
        "af",
        "affected",
        "affecting",
        "affects",
        "after",
        "afterwards",
        "ag",
        "again",
        "against",
        "ago",
        "ah",
        "ahead",
        "ai",
        "aint",
        "al",
        "all",
        "allow",
        "allows",
        "almost",
        "alone",
        "along",
        "alongside",
        "already",
        "also",
        "although",
        "always",
        "am",
        "amid",
        "amidst",
        "among",
        "amongst",
        "amoungst",
        "amount",
        "an",
        "and",
        "announce",
        "another",
        "any",
        "anybody",
        "anyhow",
        "anymore",
        "anyone",
        "anything",
        "anyway",
        "anyways",
        "anywhere",
        "ao",
        "apart",
        "apparently",
        "appear",
        "appreciate",
        "appropriate",
        "approximately",
        "aq",
        "ar",
        "are",
        "area",
        "areas",
        "aren",
        "arent",
        "arise",
        "around",
        "arpa",
        "as",
        "aside",
        "ask",
        "asked",
        "asking",
        "asks",
        "associated",
        "at",
        "au",
        "auth",
        "available",
        "aw",
        "away",
        "awfully",
        "az",
        "b",
        "ba",
        "back",
        "backed",
        "backing",
        "backs",
        "backward",
        "backwards",
        "bb",
        "bd",
        "be",
        "became",
        "because",
        "become",
        "becomes",
        "becoming",
        "been",
        "before",
        "beforehand",
        "began",
        "begin",
        "beginning",
        "beginnings",
        "begins",
        "behind",
        "being",
        "beings",
        "believe",
        "below",
        "beside",
        "besides",
        "best",
        "better",
        "between",
        "beyond",
        "bf",
        "bg",
        "bh",
        "bi",
        "big",
        "bill",
        "billion",
        "biol",
        "bj",
        "bm",
        "bn",
        "bo",
        "both",
        "bottom",
        "br",
        "brief",
        "briefly",
        "bs",
        "bt",
        "but",
        "buy",
        "bv",
        "bw",
        "by",
        "bz",
        "c",
        "ca",
        "call",
        "came",
        "can",
        "cannot",
        "cant",
        "caption",
        "case",
        "cases",
        "cause",
        "causes",
        "cc",
        "cd",
        "certain",
        "certainly",
        "cf",
        "cg",
        "ch",
        "changes",
        "ci",
        "ck",
        "cl",
        "clear",
        "clearly",
        "click",
        "cm",
        "cmon",
        "cn",
        "co",
        "co.",
        "com",
        "come",
        "comes",
        "computer",
        "con",
        "concerning",
        "consequently",
        "consider",
        "considering",
        "contain",
        "containing",
        "contains",
        "copy",
        "corresponding",
        "could",
        "couldn",
        "couldnt",
        "course",
        "cr",
        "cry",
        "cs",
        "cu",
        "currently",
        "cv",
        "cx",
        "cy",
        "cz",
        "d",
        "dare",
        "darent",
        "date",
        "de",
        "dear",
        "definitely",
        "describe",
        "described",
        "despite",
        "detail",
        "did",
        "didn",
        "didnt",
        "differ",
        "different",
        "differently",
        "directly",
        "dj",
        "dk",
        "dm",
        "do",
        "does",
        "doesn",
        "doesnt",
        "doing",
        "don",
        "done",
        "dont",
        "doubtful",
        "down",
        "downed",
        "downing",
        "downs",
        "downwards",
        "due",
        "during",
        "dz",
        "e",
        "each",
        "early",
        "ec",
        "ed",
        "edu",
        "ee",
        "effect",
        "eg",
        "eh",
        "eight",
        "eighty",
        "either",
        "eleven",
        "else",
        "elsewhere",
        "empty",
        "end",
        "ended",
        "ending",
        "ends",
        "enough",
        "entirely",
        "er",
        "es",
        "especially",
        "et",
        "et-al",
        "etc",
        "even",
        "evenly",
        "ever",
        "evermore",
        "every",
        "everybody",
        "everyone",
        "everything",
        "everywhere",
        "ex",
        "exactly",
        "example",
        "except",
        "f",
        "face",
        "faces",
        "fact",
        "facts",
        "fairly",
        "far",
        "farther",
        "felt",
        "few",
        "fewer",
        "ff",
        "fi",
        "fifteen",
        "fifth",
        "fifty",
        "fify",
        "fill",
        "find",
        "finds",
        "fire",
        "first",
        "five",
        "fix",
        "fj",
        "fk",
        "fm",
        "fo",
        "followed",
        "following",
        "follows",
        "for",
        "forever",
        "former",
        "formerly",
        "forth",
        "forty",
        "forward",
        "found",
        "four",
        "fr",
        "free",
        "from",
        "front",
        "full",
        "fully",
        "further",
        "furthered",
        "furthering",
        "furthermore",
        "furthers",
        "fx",
        "g",
        "ga",
        "gave",
        "gb",
        "gd",
        "ge",
        "general",
        "generally",
        "get",
        "gets",
        "getting",
        "gf",
        "gg",
        "gh",
        "gi",
        "give",
        "given",
        "gives",
        "giving",
        "gl",
        "gm",
        "gmt",
        "gn",
        "go",
        "goes",
        "going",
        "gone",
        "good",
        "goods",
        "got",
        "gotten",
        "gov",
        "gp",
        "gq",
        "gr",
        "great",
        "greater",
        "greatest",
        "greetings",
        "group",
        "grouped",
        "grouping",
        "groups",
        "gs",
        "gt",
        "gu",
        "gw",
        "gy",
        "h",
        "had",
        "hadnt",
        "half",
        "happens",
        "hardly",
        "has",
        "hasn",
        "hasnt",
        "have",
        "haven",
        "havent",
        "having",
        "he",
        "hed",
        "hell",
        "hello",
        "help",
        "hence",
        "her",
        "here",
        "hereafter",
        "hereby",
        "herein",
        "heres",
        "hereupon",
        "hers",
        "herself",
        "herse”",
        "hes",
        "hi",
        "hid",
        "high",
        "higher",
        "highest",
        "him",
        "himself",
        "himse”",
        "his",
        "hither",
        "hk",
        "hm",
        "hn",
        "home",
        "homepage",
        "hopefully",
        "how",
        "howbeit",
        "however",
        "hr",
        "ht",
        "htm",
        "html",
        "http",
        "hu",
        "hundred",
        "i",
        "i.e.",
        "id",
        "ie",
        "if",
        "ignored",
        "ii",
        "il",
        "ill",
        "im",
        "immediate",
        "immediately",
        "importance",
        "important",
        "in",
        "inasmuch",
        "inc",
        "inc.",
        "indeed",
        "index",
        "indicate",
        "indicated",
        "indicates",
        "information",
        "inner",
        "inside",
        "insofar",
        "instead",
        "int",
        "interest",
        "interested",
        "interesting",
        "interests",
        "into",
        "invention",
        "inward",
        "io",
        "iq",
        "ir",
        "is",
        "isn",
        "isnt",
        "it",
        "itd",
        "itll",
        "its",
        "itself",
        "itse”",
        "ive",
        "j",
        "je",
        "jm",
        "jo",
        "join",
        "jp",
        "just",
        "k",
        "ke",
        "keep",
        "keeps",
        "kept",
        "keys",
        "kg",
        "kh",
        "ki",
        "kind",
        "km",
        "kn",
        "knew",
        "know",
        "known",
        "knows",
        "kp",
        "kr",
        "kw",
        "ky",
        "kz",
        "l",
        "la",
        "large",
        "largely",
        "last",
        "lately",
        "later",
        "latest",
        "latter",
        "latterly",
        "lb",
        "lc",
        "least",
        "length",
        "less",
        "lest",
        "let",
        "lets",
        "li",
        "like",
        "liked",
        "likely",
        "likewise",
        "line",
        "little",
        "lk",
        "ll",
        "long",
        "longer",
        "longest",
        "look",
        "looking",
        "looks",
        "low",
        "lower",
        "lr",
        "ls",
        "lt",
        "ltd",
        "lu",
        "lv",
        "ly",
        "m",
        "ma",
        "made",
        "mainly",
        "make",
        "makes",
        "making",
        "man",
        "many",
        "may",
        "maybe",
        "maynt",
        "mc",
        "md",
        "me",
        "mean",
        "means",
        "meantime",
        "meanwhile",
        "member",
        "members",
        "men",
        "merely",
        "mg",
        "mh",
        "microsoft",
        "might",
        "mightnt",
        "mil",
        "mill",
        "million",
        "mine",
        "minus",
        "miss",
        "mk",
        "ml",
        "mm",
        "mn",
        "mo",
        "more",
        "moreover",
        "most",
        "mostly",
        "move",
        "mp",
        "mq",
        "mr",
        "mrs",
        "ms",
        "msie",
        "mt",
        "mu",
        "much",
        "mug",
        "must",
        "mustnt",
        "mv",
        "mw",
        "mx",
        "my",
        "myself",
        "myse”",
        "mz",
        "n",
        "na",
        "name",
        "namely",
        "nay",
        "nc",
        "nd",
        "ne",
        "near",
        "nearly",
        "necessarily",
        "necessary",
        "need",
        "needed",
        "needing",
        "neednt",
        "needs",
        "neither",
        "net",
        "netscape",
        "never",
        "neverf",
        "neverless",
        "nevertheless",
        "new",
        "newer",
        "newest",
        "next",
        "nf",
        "ng",
        "ni",
        "nine",
        "ninety",
        "nl",
        "no",
        "no-one",
        "nobody",
        "non",
        "none",
        "nonetheless",
        "noone",
        "nor",
        "normally",
        "nos",
        "not",
        "noted",
        "nothing",
        "notwithstanding",
        "novel",
        "now",
        "nowhere",
        "np",
        "nr",
        "nu",
        "null",
        "number",
        "numbers",
        "nz",
        "o",
        "obtain",
        "obtained",
        "obviously",
        "of",
        "off",
        "often",
        "oh",
        "ok",
        "okay",
        "old",
        "older",
        "oldest",
        "om",
        "omitted",
        "on",
        "once",
        "one",
        "ones",
        "only",
        "onto",
        "open",
        "opened",
        "opening",
        "opens",
        "opposite",
        "or",
        "ord",
        "order",
        "ordered",
        "ordering",
        "orders",
        "org",
        "other",
        "others",
        "otherwise",
        "ought",
        "oughtnt",
        "our",
        "ours",
        "ourselves",
        "out",
        "outside",
        "over",
        "overall",
        "owing",
        "own",
        "p",
        "pa",
        "page",
        "pages",
        "part",
        "parted",
        "particular",
        "particularly",
        "parting",
        "parts",
        "past",
        "pe",
        "per",
        "perhaps",
        "pf",
        "pg",
        "ph",
        "pk",
        "pl",
        "place",
        "placed",
        "places",
        "please",
        "plus",
        "pm",
        "pmid",
        "pn",
        "point",
        "pointed",
        "pointing",
        "points",
        "poorly",
        "possible",
        "possibly",
        "potentially",
        "pp",
        "pr",
        "predominantly",
        "present",
        "presented",
        "presenting",
        "presents",
        "presumably",
        "previously",
        "primarily",
        "probably",
        "problem",
        "problems",
        "promptly",
        "proud",
        "provided",
        "provides",
        "pt",
        "put",
        "puts",
        "pw",
        "py",
        "q",
        "qa",
        "que",
        "quickly",
        "quite",
        "qv",
        "r",
        "ran",
        "rather",
        "rd",
        "re",
        "readily",
        "really",
        "reasonably",
        "recent",
        "recently",
        "ref",
        "refs",
        "regarding",
        "regardless",
        "regards",
        "related",
        "relatively",
        "research",
        "reserved",
        "respectively",
        "resulted",
        "resulting",
        "results",
        "right",
        "ring",
        "ro",
        "room",
        "rooms",
        "round",
        "ru",
        "run",
        "rw",
        "s",
        "sa",
        "said",
        "same",
        "saw",
        "say",
        "saying",
        "says",
        "sb",
        "sc",
        "sd",
        "se",
        "sec",
        "second",
        "secondly",
        "seconds",
        "section",
        "see",
        "seeing",
        "seem",
        "seemed",
        "seeming",
        "seems",
        "seen",
        "sees",
        "self",
        "selves",
        "sensible",
        "sent",
        "serious",
        "seriously",
        "seven",
        "seventy",
        "several",
        "sg",
        "sh",
        "shall",
        "shant",
        "she",
        "shed",
        "shell",
        "shes",
        "should",
        "shouldn",
        "shouldnt",
        "show",
        "showed",
        "showing",
        "shown",
        "showns",
        "shows",
        "si",
        "side",
        "sides",
        "significant",
        "significantly",
        "similar",
        "similarly",
        "since",
        "sincere",
        "site",
        "six",
        "sixty",
        "sj",
        "sk",
        "sl",
        "slightly",
        "sm",
        "small",
        "smaller",
        "smallest",
        "sn",
        "so",
        "some",
        "somebody",
        "someday",
        "somehow",
        "someone",
        "somethan",
        "something",
        "sometime",
        "sometimes",
        "somewhat",
        "somewhere",
        "soon",
        "sorry",
        "specifically",
        "specified",
        "specify",
        "specifying",
        "sr",
        "st",
        "state",
        "states",
        "still",
        "stop",
        "strongly",
        "su",
        "sub",
        "substantially",
        "successfully",
        "such",
        "sufficiently",
        "suggest",
        "sup",
        "sure",
        "sv",
        "sy",
        "system",
        "sz",
        "t",
        "take",
        "taken",
        "taking",
        "tc",
        "td",
        "tell",
        "ten",
        "tends",
        "test",
        "text",
        "tf",
        "tg",
        "th",
        "than",
        "thank",
        "thanks",
        "thanx",
        "that",
        "thatll",
        "thats",
        "thatve",
        "the",
        "their",
        "theirs",
        "them",
        "themselves",
        "then",
        "thence",
        "there",
        "thereafter",
        "thereby",
        "thered",
        "therefore",
        "therein",
        "therell",
        "thereof",
        "therere",
        "theres",
        "thereto",
        "thereupon",
        "thereve",
        "these",
        "they",
        "theyd",
        "theyll",
        "theyre",
        "theyve",
        "thick",
        "thin",
        "thing",
        "things",
        "think",
        "thinks",
        "third",
        "thirty",
        "this",
        "thorough",
        "thoroughly",
        "those",
        "thou",
        "though",
        "thoughh",
        "thought",
        "thoughts",
        "thousand",
        "three",
        "throug",
        "through",
        "throughout",
        "thru",
        "thus",
        "til",
        "till",
        "tip",
        "tis",
        "tj",
        "tk",
        "tm",
        "tn",
        "to",
        "today",
        "together",
        "too",
        "took",
        "top",
        "toward",
        "towards",
        "tp",
        "tr",
        "tried",
        "tries",
        "trillion",
        "truly",
        "try",
        "trying",
        "ts",
        "tt",
        "turn",
        "turned",
        "turning",
        "turns",
        "tv",
        "tw",
        "twas",
        "twelve",
        "twenty",
        "twice",
        "two",
        "tz",
        "u",
        "ua",
        "ug",
        "uk",
        "um",
        "un",
        "under",
        "underneath",
        "undoing",
        "unfortunately",
        "unless",
        "unlike",
        "unlikely",
        "until",
        "unto",
        "up",
        "upon",
        "ups",
        "upwards",
        "us",
        "use",
        "used",
        "useful",
        "usefully",
        "usefulness",
        "uses",
        "using",
        "usually",
        "uucp",
        "uy",
        "uz",
        "v",
        "va",
        "value",
        "various",
        "vc",
        "ve",
        "versus",
        "very",
        "vg",
        "vi",
        "via",
        "viz",
        "vn",
        "vol",
        "vols",
        "vs",
        "vu",
        "w",
        "want",
        "wanted",
        "wanting",
        "wants",
        "was",
        "wasn",
        "wasnt",
        "way",
        "ways",
        "we",
        "web",
        "webpage",
        "website",
        "wed",
        "welcome",
        "well",
        "wells",
        "went",
        "were",
        "weren",
        "werent",
        "weve",
        "wf",
        "what",
        "whatever",
        "whatll",
        "whats",
        "whatve",
        "when",
        "whence",
        "whenever",
        "where",
        "whereafter",
        "whereas",
        "whereby",
        "wherein",
        "wheres",
        "whereupon",
        "wherever",
        "whether",
        "which",
        "whichever",
        "while",
        "whilst",
        "whim",
        "whither",
        "who",
        "whod",
        "whoever",
        "whole",
        "wholl",
        "whom",
        "whomever",
        "whos",
        "whose",
        "why",
        "widely",
        "width",
        "will",
        "willing",
        "wish",
        "with",
        "within",
        "without",
        "won",
        "wonder",
        "wont",
        "words",
        "work",
        "worked",
        "working",
        "works",
        "world",
        "would",
        "wouldn",
        "wouldnt",
        "ws",
        "www",
        "x",
        "y",
        "ye",
        "year",
        "years",
        "yes",
        "yet",
        "you",
        "youd",
        "youll",
        "young",
        "younger",
        "youngest",
        "your",
        "youre",
        "yours",
        "yourself",
        "yourselves",
        "youve",
        "yt",
        "yu",
        "z",
        "za",
        "zero",
        "zm",
        "zr",
    ]

    # Combine all categories including the new ones
    all_stopwords = (
        units
        + prefixes
        + generic_technical
        + atlas_generic
        + math_stats
        + document_structure
        + releases
        + years
        + things
        + equation_leftovers
        + version_patterns
        + computing_terms
        + size_speed_units
        + infrastructure_terms
        + status_terms
        + action_terms
        + file_terms
        + time_terms
        + additional_technical
        + pure_numbers
        + build_terms
        + detector_terms
        + doc_patterns
        + time_date
        + numerical_patterns
        + file_paths
        + status_terms
        + file_types
        + other_english_stop_words
        + github_english_stop_words
    )

    # Add common variations (upper/lower/capitalized)
    final_stopwords = set()
    for word in all_stopwords:
        final_stopwords.add(word.lower())
        final_stopwords.add(word.upper())
        final_stopwords.add(word.capitalize())
        # Add common plural forms for relevant words
        if not word.endswith("s") and not word.endswith("S"):
            final_stopwords.add(f"{word.lower()}s")
            final_stopwords.add(f"{word.upper()}s")
            final_stopwords.add(f"{word.capitalize()}s")

    return list(final_stopwords)
